"""Define the pynecone state specification."""
from __future__ import annotations

import asyncio
import functools
import traceback
from abc import ABC
from collections import defaultdict
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Type,
    Union,
)

import cloudpickle
import pydantic
from redis import Redis

from pynecone import constants
from pynecone.base import Base
from pynecone.event import Event, EventHandler, fix_events, window_alert
from pynecone.utils import format, prerequisites, types
from pynecone.var import BaseVar, ComputedVar, PCDict, PCList, Var

Delta = Dict[str, Any]


class State(Base, ABC, extra=pydantic.Extra.allow):
    """The state of the app."""

    # A map from the var name to the var.
    vars: ClassVar[Dict[str, Var]] = {}

    # The base vars of the class.
    base_vars: ClassVar[Dict[str, BaseVar]] = {}

    # The computed vars of the class.
    computed_vars: ClassVar[Dict[str, ComputedVar]] = {}

    # Vars inherited by the parent state.
    inherited_vars: ClassVar[Dict[str, Var]] = {}

    # Backend vars that are never sent to the client.
    backend_vars: ClassVar[Dict[str, Any]] = {}

    # Backend vars inherited
    inherited_backend_vars: ClassVar[Dict[str, Any]] = {}

    # The event handlers.
    event_handlers: ClassVar[Dict[str, EventHandler]] = {}

    # The parent state.
    parent_state: Optional[State] = None

    # The substates of the state.
    substates: Dict[str, State] = {}

    # The set of dirty vars.
    dirty_vars: Set[str] = set()

    # The set of dirty substates.
    dirty_substates: Set[str] = set()

    # The routing path that triggered the state
    router_data: Dict[str, Any] = {}

    # Mapping of var name to set of computed variables that depend on it
    computed_var_dependencies: Dict[str, Set[str]] = {}

    # Whether to track accessed vars.
    track_vars: bool = False

    # The current set of accessed vars during tracking.
    tracked_vars: Set[str] = set()

    def __init__(self, *args, parent_state: Optional[State] = None, **kwargs):
        """Initialize the state.

        Args:
            *args: The args to pass to the Pydantic init method.
            parent_state: The parent state.
            **kwargs: The kwargs to pass to the Pydantic init method.
        """
        kwargs["parent_state"] = parent_state
        super().__init__(*args, **kwargs)

        # Setup the substates.
        for substate in self.get_substates():
            self.substates[substate.get_name()] = substate(parent_state=self)

        # Convert the event handlers to functions.
        for name, event_handler in self.event_handlers.items():
            setattr(self, name, event_handler.fn)

        # Initialize the mutable fields.
        self._init_mutable_fields()

        # Initialize computed vars dependencies.
        self.computed_var_dependencies = defaultdict(set)
        for cvar in self.computed_vars:
            self.tracked_vars = set()

            # Enable tracking and get the computed var.
            self.track_vars = True
            self.__getattribute__(cvar)
            self.track_vars = False

            # Add the dependencies.
            for var in self.tracked_vars:
                self.computed_var_dependencies[var].add(cvar)

    def _init_mutable_fields(self):
        """Initialize mutable fields.

        So that mutation to them can be detected by the app:
        * list
        """
        for field in self.base_vars.values():
            value = getattr(self, field.name)

            value_in_pc_data = _convert_mutable_datatypes(
                value, self._reassign_field, field.name
            )

            if types._issubclass(field.type_, Union[List, Dict]):
                setattr(self, field.name, value_in_pc_data)

        self.clean()

    def _reassign_field(self, field_name: str):
        """Reassign the given field.

        Primarily for mutation in fields of mutable data types.

        Args:
            field_name: The name of the field we want to reassign
        """
        setattr(
            self,
            field_name,
            getattr(self, field_name),
        )

    def __repr__(self) -> str:
        """Get the string representation of the state.

        Returns:
            The string representation of the state.
        """
        return f"{self.__class__.__name__}({self.dict()})"

    @classmethod
    def __init_subclass__(cls, **kwargs):
        """Do some magic for the subclass initialization.

        Args:
            **kwargs: The kwargs to pass to the pydantic init_subclass method.
        """
        super().__init_subclass__(**kwargs)

        # Get the parent vars.
        parent_state = cls.get_parent_state()
        if parent_state is not None:
            cls.inherited_vars = parent_state.vars
            cls.inherited_backend_vars = parent_state.backend_vars

        cls.new_backend_vars = {
            name: value
            for name, value in cls.__dict__.items()
            if types.is_backend_variable(name)
            and name not in cls.inherited_backend_vars
        }

        cls.backend_vars = {**cls.inherited_backend_vars, **cls.new_backend_vars}

        # Set the base and computed vars.
        cls.base_vars = {
            f.name: BaseVar(name=f.name, type_=f.outer_type_).set_state(cls)
            for f in cls.get_fields().values()
            if f.name not in cls.get_skip_vars()
        }
        cls.computed_vars = {
            v.name: v.set_state(cls)
            for v in cls.__dict__.values()
            if isinstance(v, ComputedVar)
        }
        cls.vars = {
            **cls.inherited_vars,
            **cls.base_vars,
            **cls.computed_vars,
        }
        cls.computed_var_dependencies = {}
        cls.event_handlers = {}

        # Setup the base vars at the class level.
        for prop in cls.base_vars.values():
            cls._init_var(prop)

        # Set up the event handlers.
        events = {
            name: fn
            for name, fn in cls.__dict__.items()
            if not name.startswith("_")
            and isinstance(fn, Callable)
            and not isinstance(fn, EventHandler)
        }
        for name, fn in events.items():
            handler = EventHandler(fn=fn)
            cls.event_handlers[name] = handler
            setattr(cls, name, handler)

    @classmethod
    def get_skip_vars(cls) -> Set[str]:
        """Get the vars to skip when serializing.

        Returns:
            The vars to skip when serializing.
        """
        return set(cls.inherited_vars) | {
            "parent_state",
            "substates",
            "dirty_vars",
            "dirty_substates",
            "router_data",
            "computed_var_dependencies",
            "track_vars",
            "tracked_vars",
        }

    @classmethod
    @functools.lru_cache()
    def get_parent_state(cls) -> Optional[Type[State]]:
        """Get the parent state.

        Returns:
            The parent state.
        """
        parent_states = [
            base
            for base in cls.__bases__
            if types._issubclass(base, State) and base is not State
        ]
        assert len(parent_states) < 2, "Only one parent state is allowed."
        return parent_states[0] if len(parent_states) == 1 else None  # type: ignore

    @classmethod
    @functools.lru_cache()
    def get_substates(cls) -> Set[Type[State]]:
        """Get the substates of the state.

        Returns:
            The substates of the state.
        """
        return set(cls.__subclasses__())

    @classmethod
    @functools.lru_cache()
    def get_name(cls) -> str:
        """Get the name of the state.

        Returns:
            The name of the state.
        """
        return format.to_snake_case(cls.__name__)

    @classmethod
    @functools.lru_cache()
    def get_full_name(cls) -> str:
        """Get the full name of the state.

        Returns:
            The full name of the state.
        """
        name = cls.get_name()
        parent_state = cls.get_parent_state()
        if parent_state is not None:
            name = ".".join((parent_state.get_full_name(), name))
        return name

    @classmethod
    @functools.lru_cache()
    def get_class_substate(cls, path: Sequence[str]) -> Type[State]:
        """Get the class substate.

        Args:
            path: The path to the substate.

        Returns:
            The class substate.

        Raises:
            ValueError: If the substate is not found.
        """
        if len(path) == 0:
            return cls
        if path[0] == cls.get_name():
            if len(path) == 1:
                return cls
            path = path[1:]
        for substate in cls.get_substates():
            if path[0] == substate.get_name():
                return substate.get_class_substate(path[1:])
        raise ValueError(f"Invalid path: {path}")

    @classmethod
    def get_class_var(cls, path: Sequence[str]) -> Any:
        """Get the class var.

        Args:
            path: The path to the var.

        Returns:
            The class var.

        Raises:
            ValueError: If the path is invalid.
        """
        path, name = path[:-1], path[-1]
        substate = cls.get_class_substate(tuple(path))
        if not hasattr(substate, name):
            raise ValueError(f"Invalid path: {path}")
        return getattr(substate, name)

    @classmethod
    def _init_var(cls, prop: BaseVar):
        """Initialize a variable.

        Args:
            prop (BaseVar): The variable to initialize

        Raises:
            TypeError: if the variable has an incorrect type
        """
        if not types.is_valid_var_type(prop.type_):
            raise TypeError(
                "State vars must be primitive Python types, "
                "Plotly figures, Pandas dataframes, "
                "or subclasses of pc.Base. "
                f'Found var "{prop.name}" with type {prop.type_}.'
            )
        cls._set_var(prop)
        cls._create_setter(prop)
        cls._set_default_value(prop)

    @classmethod
    def add_var(cls, name: str, type_: Any, default_value: Any = None):
        """Add dynamically a variable to the State.

        The variable added this way can be used in the same way as a variable
        defined statically in the model.

        Args:
            name: The name of the variable
            type_: The type of the variable
            default_value: The default value of the variable

        Raises:
            NameError: if a variable of this name already exists
        """
        if name in cls.__fields__:
            raise NameError(
                f"The variable '{name}' already exist. Use a different name"
            )

        # create the variable based on name and type
        var = BaseVar(name=name, type_=type_)
        var.set_state(cls)

        # add the pydantic field dynamically (must be done before _init_var)
        cls.add_field(var, default_value)

        cls._init_var(var)

        # update the internal dicts so the new variable is correctly handled
        cls.base_vars.update({name: var})
        cls.vars.update({name: var})

    @classmethod
    def _set_var(cls, prop: BaseVar):
        """Set the var as a class member.

        Args:
            prop: The var instance to set.
        """
        setattr(cls, prop.name, prop)

    @classmethod
    def _create_setter(cls, prop: BaseVar):
        """Create a setter for the var.

        Args:
            prop: The var to create a setter for.
        """
        setter_name = prop.get_setter_name(include_state=False)
        if setter_name not in cls.__dict__:
            event_handler = EventHandler(fn=prop.get_setter())
            cls.event_handlers[setter_name] = event_handler
            setattr(cls, setter_name, event_handler)

    @classmethod
    def _set_default_value(cls, prop: BaseVar):
        """Set the default value for the var.

        Args:
            prop: The var to set the default value for.
        """
        # Get the pydantic field for the var.
        field = cls.get_fields()[prop.name]
        default_value = prop.get_default_value()
        if field.required and default_value is not None:
            field.required = False
            field.default = default_value

    def get_token(self) -> str:
        """Return the token of the client associated with this state.

        Returns:
            The token of the client.
        """
        return self.router_data.get(constants.RouteVar.CLIENT_TOKEN, "")

    def get_sid(self) -> str:
        """Return the session ID of the client associated with this state.

        Returns:
            The session ID of the client.
        """
        return self.router_data.get(constants.RouteVar.SESSION_ID, "")

    def get_headers(self) -> Dict:
        """Return the headers of the client associated with this state.

        Returns:
            The headers of the client.
        """
        return self.router_data.get(constants.RouteVar.HEADERS, {})

    def get_client_ip(self) -> str:
        """Return the IP of the client associated with this state.

        Returns:
            The IP of the client.
        """
        return self.router_data.get(constants.RouteVar.CLIENT_IP, "")

    def get_current_page(self) -> str:
        """Obtain the path of current page from the router data.

        Returns:
            The current page.
        """
        return self.router_data.get(constants.RouteVar.PATH, "")

    def get_query_params(self) -> Dict[str, str]:
        """Obtain the query parameters for the queried page.

        The query object contains both the URI parameters and the GET parameters.

        Returns:
            The dict of query parameters.
        """
        return self.router_data.get(constants.RouteVar.QUERY, {})

    @classmethod
    def setup_dynamic_args(cls, args: dict[str, str]):
        """Set up args for easy access in renderer.

        Args:
            args: a dict of args
        """

        def argsingle_factory(param):
            @ComputedVar
            def inner_func(self) -> str:
                return self.get_query_params().get(param, "")

            return inner_func

        def arglist_factory(param):
            @ComputedVar
            def inner_func(self) -> List:
                return self.get_query_params().get(param, [])

            return inner_func

        for param, value in args.items():
            if value == constants.RouteArgType.SINGLE:
                func = argsingle_factory(param)
            elif value == constants.RouteArgType.LIST:
                func = arglist_factory(param)
            else:
                continue
            cls.computed_vars[param] = func.set_state(cls)  # type: ignore
            setattr(cls, param, func)

    def __getattribute__(self, name: str) -> Any:
        """Get the state var.

        If the var is inherited, get the var from the parent state.

        If the Var is a dependent of a ComputedVar, track this status in computed_var_dependencies.

        Args:
            name: The name of the var.

        Returns:
            The value of the var.
        """
        # If the state hasn't been initialized yet, return the default value.
        if not super().__getattribute__("__dict__"):
            return super().__getattribute__(name)

        # Check if tracking is enabled.
        if super().__getattribute__("track_vars"):
            # Get the non-computed vars.
            all_vars = {
                **super().__getattribute__("vars"),
                **super().__getattribute__("backend_vars"),
            }
            # Add the var to the tracked vars.
            if name in all_vars:
                super().__getattribute__("tracked_vars").add(name)

        inherited_vars = {
            **super().__getattribute__("inherited_vars"),
            **super().__getattribute__("inherited_backend_vars"),
        }
        if name in inherited_vars:
            return getattr(super().__getattribute__("parent_state"), name)
        elif name in super().__getattribute__("backend_vars"):
            return super().__getattribute__("backend_vars").__getitem__(name)
        return super().__getattribute__(name)

    def __setattr__(self, name: str, value: Any):
        """Set the attribute.

        If the attribute is inherited, set the attribute on the parent state.

        Args:
            name: The name of the attribute.
            value: The value of the attribute.
        """
        # Set the var on the parent state.
        inherited_vars = {**self.inherited_vars, **self.inherited_backend_vars}
        if name in inherited_vars:
            setattr(self.parent_state, name, value)
            return

        if types.is_backend_variable(name):
            self.backend_vars.__setitem__(name, value)
            self.dirty_vars.add(name)
            self.mark_dirty()
            return

        # Set the attribute.
        super().__setattr__(name, value)

        # Add the var to the dirty list.
        if name in self.vars:
            self.dirty_vars.add(name)
            self.mark_dirty()

    def reset(self):
        """Reset all the base vars to their default values."""
        # Reset the base vars.
        fields = self.get_fields()
        for prop_name in self.base_vars:
            setattr(self, prop_name, fields[prop_name].default)

        # Recursively reset the substates.
        for substate in self.substates.values():
            substate.reset()

        # Clean the state.
        self.clean()

    def get_substate(self, path: Sequence[str]) -> Optional[State]:
        """Get the substate.

        Args:
            path: The path to the substate.

        Returns:
            The substate.

        Raises:
            ValueError: If the substate is not found.
        """
        if len(path) == 0:
            return self
        if path[0] == self.get_name():
            if len(path) == 1:
                return self
            path = path[1:]
        if path[0] not in self.substates:
            raise ValueError(f"Invalid path: {path}")
        return self.substates[path[0]].get_substate(path[1:])

    async def _process(self, event: Event) -> StateUpdate:
        """Obtain event info and process event.

        Args:
            event: The event to process.

        Returns:
            The state update after processing the event.

        Raises:
            ValueError: If the state value is None.
        """
        # Get the event handler.
        path = event.name.split(".")
        path, name = path[:-1], path[-1]
        substate = self.get_substate(path)
        handler = substate.event_handlers[name]  # type: ignore

        if not substate:
            raise ValueError(
                "The value of state cannot be None when processing an event."
            )

        return await self._process_event(
            handler=handler,
            state=substate,
            payload=event.payload,
            token=event.token,
        )

    async def _process_event(
        self, handler: EventHandler, state: State, payload: Dict, token: str
    ) -> StateUpdate:
        """Process event.

        Args:
            handler: Eventhandler to process.
            state: State to process the handler.
            payload: The event payload.
            token: Client token.

        Returns:
            The state update after processing the event.
        """
        fn = functools.partial(handler.fn, state)
        try:
            if asyncio.iscoroutinefunction(fn.func):
                events = await fn(**payload)
            else:
                events = fn(**payload)
        except Exception:
            error = traceback.format_exc()
            print(error)
            events = fix_events(
                [window_alert("An error occurred. See logs for details.")], token
            )
            return StateUpdate(events=events)

        # Fix the returned events.
        events = fix_events(events, token)

        # Get the delta after processing the event.
        delta = self.get_delta()

        # Reset the dirty vars.
        self.clean()

        # Return the state update.
        return StateUpdate(delta=delta, events=events)

    def _dirty_computed_vars(
        self, from_vars: Optional[Set[str]] = None, check: bool = False
    ) -> Set[str]:
        """Get ComputedVars that need to be recomputed based on dirty_vars.

        Args:
            from_vars: find ComputedVar that depend on this set of vars. If unspecified, will use the dirty_vars.
            check: Whether to perform the check.

        Returns:
            Set of computed vars to include in the delta.
        """
        # If checking is disabled, return all computed vars.
        if not check:
            return set(self.computed_vars)

        # Return only the computed vars that depend on the dirty vars.
        return set(
            cvar
            for dirty_var in from_vars or self.dirty_vars
            for cvar in self.computed_vars
            if cvar in self.computed_var_dependencies.get(dirty_var, set())
        )

    def get_delta(self, check: bool = False) -> Delta:
        """Get the delta for the state.

        Args:
            check: Whether to check for dirty computed vars.

        Returns:
            The delta for the state.
        """
        delta = {}

        # Return the dirty vars, as well as computed vars depending on dirty vars.
        subdelta = {
            prop: getattr(self, prop)
            for prop in self.dirty_vars | self._dirty_computed_vars(check=check)
            if not types.is_backend_variable(prop)
        }
        if len(subdelta) > 0:
            delta[self.get_full_name()] = subdelta

        # Recursively find the substate deltas.
        substates = self.substates
        for substate in self.dirty_substates:
            delta.update(substates[substate].get_delta())

        # Format the delta.
        delta = format.format_state(delta)

        # Return the delta.
        return delta

    def mark_dirty(self):
        """Mark the substate and all parent states as dirty."""
        if self.parent_state is not None:
            self.parent_state.dirty_substates.add(self.get_name())
            self.parent_state.mark_dirty()

    def clean(self):
        """Reset the dirty vars."""
        # Recursively clean the substates.
        for substate in self.dirty_substates:
            self.substates[substate].clean()

        # Clean this state.
        self.dirty_vars = set()
        self.dirty_substates = set()

    def dict(self, include_computed: bool = True, **kwargs) -> Dict[str, Any]:
        """Convert the object to a dictionary.

        Args:
            include_computed: Whether to include computed vars.
            **kwargs: Kwargs to pass to the pydantic dict method.

        Returns:
            The object as a dictionary.
        """
        base_vars = {
            prop_name: self.get_value(getattr(self, prop_name))
            for prop_name in self.base_vars
        }
        computed_vars = (
            {
                # Include the computed vars.
                prop_name: self.get_value(getattr(self, prop_name))
                for prop_name in self.computed_vars
            }
            if include_computed
            else {}
        )
        substate_vars = {
            k: v.dict(include_computed=include_computed, **kwargs)
            for k, v in self.substates.items()
        }
        variables = {**base_vars, **computed_vars, **substate_vars}
        return {k: variables[k] for k in sorted(variables)}


class DefaultState(State):
    """The default empty state."""

    pass


class StateUpdate(Base):
    """A state update sent to the frontend."""

    # The state delta.
    delta: Delta = {}

    # Events to be added to the event queue.
    events: List[Event] = []


class StateManager(Base):
    """A class to manage many client states."""

    # The state class to use.
    state: Type[State] = DefaultState

    # The mapping of client ids to states.
    states: Dict[str, State] = {}

    # The token expiration time (s).
    token_expiration: int = constants.TOKEN_EXPIRATION

    # The redis client to use.
    redis: Optional[Redis] = None

    def setup(self, state: Type[State]):
        """Set up the state manager.

        Args:
            state: The state class to use.
        """
        self.state = state
        self.redis = prerequisites.get_redis()

    def get_state(self, token: str) -> State:
        """Get the state for a token.

        Args:
            token: The token to get the state for.

        Returns:
            The state for the token.
        """
        if self.redis is not None:
            redis_state = self.redis.get(token)
            if redis_state is None:
                self.set_state(token, self.state())
                return self.get_state(token)
            return cloudpickle.loads(redis_state)

        if token not in self.states:
            self.states[token] = self.state()
        return self.states[token]

    def set_state(self, token: str, state: State):
        """Set the state for a token.

        Args:
            token: The token to set the state for.
            state: The state to set.
        """
        if self.redis is None:
            return
        self.redis.set(token, cloudpickle.dumps(state), ex=self.token_expiration)


def _convert_mutable_datatypes(
    field_value: Any, reassign_field: Callable, field_name: str
) -> Any:
    """Recursively convert mutable data to the Pc data types.

    Note: right now only list & dict would be handled recursively.

    Args:
        field_value: The target field_value.
        reassign_field:
            The function to reassign the field in the parent state.
        field_name: the name of the field in the parent state

    Returns:
        The converted field_value
    """
    if isinstance(field_value, list):
        for index in range(len(field_value)):
            field_value[index] = _convert_mutable_datatypes(
                field_value[index], reassign_field, field_name
            )

        field_value = PCList(
            field_value, reassign_field=reassign_field, field_name=field_name
        )

    if isinstance(field_value, dict):
        for key, value in field_value.items():
            field_value[key] = _convert_mutable_datatypes(
                value, reassign_field, field_name
            )
        field_value = PCDict(
            field_value, reassign_field=reassign_field, field_name=field_name
        )
    return field_value



"""A link component."""

from pynecone.components.component import Component
from pynecone.components.libs.chakra import ChakraComponent
from pynecone.components.navigation.nextlink import NextLink
from pynecone.var import Var


class Link(ChakraComponent):
    """Link to another page."""

    tag = "Link"

    # The rel.
    rel: Var[str]

    # The page to link to.
    href: Var[str]

    # The text to display.
    text: Var[str]

    # If true, the link will open in new tab.
    is_external: Var[bool]

    @classmethod
    def create(cls, *children, **props) -> Component:
        """Create a NextJS link component, wrapping a Chakra link component.

        Args:
            *children: The children to pass to the component.
            **props: The attributes to pass to the component.

        Returns:
            The component.
        """
        kwargs = {"href": props.pop("href") if "href" in props else "#"}
        return NextLink.create(super().create(*children, **props), **kwargs)



"""HTML constants."""

# See https://html.spec.whatwg.org/#elements-3.
ELEMENTS = {
    "a",
    "abbr",
    "address",
    "area",
    "article",
    "aside",
    "audio",
    "b",
    "base",
    "bdi",
    "bdo",
    "blockquote",
    "body",
    "br",
    "button",
    "canvas",
    "caption",
    "cite",
    "code",
    "col",
    "colgroup",
    "data",
    "datalist",
    "dd",
    "del",
    "details",
    "dfn",
    "dialog",
    "div",
    "dl",
    "dt",
    "em",
    "embed",
    "fieldset",
    "figcaption",
    "figure",
    "footer",
    "form",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "head",
    "header",
    "hr",
    "html",
    "i",
    "iframe",
    "img",
    "input",
    "ins",
    "kbd",
    "label",
    "legend",
    "li",
    "link",
    "main",
    "map",
    "mark",
    "math",
    "menu",
    "meta",
    "meter",
    "nav",
    "noscript",
    "object",
    "ol",
    "optgroup",
    "option",
    "output",
    "p",
    "picture",
    "portal",
    "pre",
    "progress",
    "q",
    "rp",
    "rt",
    "ruby",
    "s",
    "samp",
    "script",
    "section",
    "select",
    "slot",
    "small",
    "source",
    "span",
    "strong",
    "style",
    "sub",
    "summary",
    "sup",
    "svg",
    "table",
    "tbody",
    "td",
    "template",
    "textarea",
    "tfoot",
    "th",
    "thead",
    "time",
    "title",
    "tr",
    "track",
    "u",
    "ul",
    "var",
    "video",
    "wbr",
}

# See https://developer.mozilla.org/en-US/docs/Web/HTML/Attributes.
ATTR_TO_ELEMENTS = {
    "accept": {"form", "input"},
    "accept-charset": {"form"},
    "accesskey": ELEMENTS,
    "action": {"form"},
    "align": {
        "applet",
        "caption",
        "col",
        "colgroup",
        "hr",
        "iframe",
        "img",
        "table",
        "tbody",
        "td",
        "tfoot",
        "th",
        "thead",
        "tr",
    },
    "allow": {"iframe"},
    "alt": {"applet", "area", "img", "input"},
    "async": {"script"},
    "autocapitalize": ELEMENTS,
    "autocomplete": {"form", "input", "select", "textarea"},
    "autofocus": {"button", "input", "keygen", "select", "textarea"},
    "autoplay": {"audio", "video"},
    "background": {"body", "table", "td", "th"},
    "bgcolor": {
        "body",
        "col",
        "colgroup",
        "marquee",
        "table",
        "tbody",
        "tfoot",
        "td",
        "th",
        "tr",
    },
    "border": {"img", "object", "table"},
    "buffered": {"audio", "video"},
    "capture": {"input"},
    "challenge": {"keygen"},
    "charset": {"meta", "script"},
    "checked": {"input"},
    "cite": {"blockquote", "del", "ins", "q"},
    "class": ELEMENTS,
    "code": {"applet"},
    "codebase": {"applet"},
    "color": {"font", "hr"},
    "cols": {"textarea"},
    "colspan": {"td", "th"},
    "content": {"meta"},
    "contenteditable": ELEMENTS,
    "contextmenu": ELEMENTS,
    "controls": {"audio", "video"},
    "coords": {"area"},
    "crossorigin": {"audio", "img", "link", "script", "video"},
    "csp": {"iframe"},
    "data": {"object"},
    "datetime": {"del", "ins", "time"},
    "decoding": {"img"},
    "default": {"track"},
    "defer": {"script"},
    "dir": ELEMENTS,
    "dirname": {"input", "textarea"},
    "disabled": {
        "button",
        "fieldset",
        "input",
        "keygen",
        "optgroup",
        "option",
        "select",
        "textarea",
    },
    "download": {"a", "area"},
    "draggable": ELEMENTS,
    "enctype": {"form"},
    "enterkeyhint": ELEMENTS,
    "for": {"label", "output"},
    "form": {
        "button",
        "fieldset",
        "input",
        "keygen",
        "label",
        "meter",
        "object",
        "output",
        "progress",
        "select",
        "textarea",
    },
    "formaction": {"button", "input"},
    "formenctype": {"button", "input"},
    "formmethod": {"button", "input"},
    "formnovalidate": {"button", "input"},
    "formtarget": {"button", "input"},
    "headers": {"td", "th"},
    "height": {"canvas", "embed", "iframe", "img", "input", "object", "video"},
    "hidden": ELEMENTS,
    "high": {"meter"},
    "href": {"a", "area", "base", "link"},
    "hreflang": {"a", "area", "link"},
    "http-equiv": {"meta"},
    "icon": {"command"},
    "id": ELEMENTS,
    "integrity": {"link", "script"},
    "intrinsicsize": {"img"},
    "inputmode": ELEMENTS,
    "ismap": {"img"},
    "itemprop": ELEMENTS,
    "keytype": {"keygen"},
    "kind": {"track"},
    "label": {"optgroup", "option", "track"},
    "lang": ELEMENTS,
    "language": {"script"},
    "loading": {"img", "iframe"},
    "list": {"input"},
    "loop": {"audio", "bgsound", "marquee", "video"},
    "low": {"meter"},
    "manifest": {"html"},
    "max": {"input", "meter", "progress"},
    "maxlength": {"input", "textarea"},
    "minlength": {"input", "textarea"},
    "media": {"a", "area", "link", "source", "style"},
    "method": {"form"},
    "min": {"input", "meter"},
    "multiple": {"input", "select"},
    "muted": {"audio", "video"},
    "name": {
        "button",
        "form",
        "fieldset",
        "iframe",
        "input",
        "keygen",
        "object",
        "output",
        "select",
        "textarea",
        "map",
        "meta",
        "param",
    },
    "novalidate": {"form"},
    "open": {"details", "dialog"},
    "optimum": {"meter"},
    "pattern": {"input"},
    "ping": {"a", "area"},
    "placeholder": {"input", "textarea"},
    "playsinline": {"video"},
    "poster": {"video"},
    "preload": {"audio", "video"},
    "readonly": {"input", "textarea"},
    "referrerpolicy": {"a", "area", "iframe", "img", "link", "script"},
    "rel": {"a", "area", "link"},
    "required": {"input", "select", "textarea"},
    "reversed": {"ol"},
    "role": ELEMENTS,
    "rows": {"textarea"},
    "rowspan": {"td", "th"},
    "sandbox": {"iframe"},
    "scope": {"th"},
    "scoped": {"style"},
    "selected": {"option"},
    "shape": {"a", "area"},
    "size": {"input", "select"},
    "sizes": {"link", "img", "source"},
    "slot": ELEMENTS,
    "span": {"col", "colgroup"},
    "spellcheck": ELEMENTS,
    "src": {
        "audio",
        "embed",
        "iframe",
        "img",
        "input",
        "script",
        "source",
        "track",
        "video",
    },
    "srcdoc": {"iframe"},
    "srclang": {"track"},
    "srcset": {"img", "source"},
    "start": {"ol"},
    "step": {"input"},
    "style": ELEMENTS,
    "summary": {"table"},
    "tabindex": ELEMENTS,
    "target": {"a", "area", "base", "form"},
    "title": ELEMENTS,
    "translate": ELEMENTS,
    "type": {
        "button",
        "input",
        "embed",
        "object",
        "ol",
        "script",
        "source",
        "style",
        "menu",
        "link",
    },
    "usemap": {"img", "input", "object"},
    "value": {
        "button",
        "data",
        "input",
        "li",
        "meter",
        "option",
        "progress",
        "param",
    },
    "width": {"canvas", "embed", "iframe", "img", "input", "object", "video"},
    "wrap": {"textarea"},
}



"""Path operations."""

from __future__ import annotations

import os
import shutil
from typing import Optional

# Shorthand for join.
join = os.linesep.join


def rm(path: str):
    """Remove a file or directory.

    Args:
        path: The path to the file or directory.
    """
    if os.path.isdir(path):
        shutil.rmtree(path)
    elif os.path.isfile(path):
        os.remove(path)


def cp(src: str, dest: str, overwrite: bool = True) -> bool:
    """Copy a file or directory.

    Args:
        src: The path to the file or directory.
        dest: The path to the destination.
        overwrite: Whether to overwrite the destination.

    Returns:
        Whether the copy was successful.
    """
    if src == dest:
        return False
    if not overwrite and os.path.exists(dest):
        return False
    if os.path.isdir(src):
        rm(dest)
        shutil.copytree(src, dest)
    else:
        shutil.copyfile(src, dest)
    return True


def mv(src: str, dest: str, overwrite: bool = True) -> bool:
    """Move a file or directory.

    Args:
        src: The path to the file or directory.
        dest: The path to the destination.
        overwrite: Whether to overwrite the destination.

    Returns:
        Whether the move was successful.
    """
    if src == dest:
        return False
    if not overwrite and os.path.exists(dest):
        return False
    rm(dest)
    shutil.move(src, dest)
    return True


def mkdir(path: str):
    """Create a directory.

    Args:
        path: The path to the directory.
    """
    if not os.path.exists(path):
        os.makedirs(path)


def ln(src: str, dest: str, overwrite: bool = False) -> bool:
    """Create a symbolic link.

    Args:
        src: The path to the file or directory.
        dest: The path to the destination.
        overwrite: Whether to overwrite the destination.

    Returns:
        Whether the link was successful.
    """
    if src == dest:
        return False
    if not overwrite and (os.path.exists(dest) or os.path.islink(dest)):
        return False
    if os.path.isdir(src):
        rm(dest)
        os.symlink(src, dest, target_is_directory=True)
    else:
        os.symlink(src, dest)
    return True


def which(program: str) -> Optional[str]:
    """Find the path to an executable.

    Args:
        program: The name of the executable.

    Returns:
        The path to the executable.
    """
    return shutil.which(program)



"""Import all classes and functions the end user will need to make an app.

Anything imported here will be available in the default Pynecone import as `pc.*`.
"""

from . import el
from .app import App, UploadFile
from .base import Base
from .components import *
from .components.component import custom_component as memo
from .components.graphing.victory import data
from .config import Config, DBConfig
from .constants import Env, Transports
from .event import (
    EVENT_ARG,
    EventChain,
    console_log,
    redirect,
    window_alert,
)
from .event import FileUpload as upload_files
from .middleware import Middleware
from .model import Model, session
from .route import route
from .state import ComputedVar as var
from .state import State
from .style import toggle_color_mode
from .var import Var



"""Responsive components."""

from pynecone.components.layout.box import Box


# Add responsive styles shortcuts.
def mobile_only(*children, **props):
    """Create a component that is only visible on mobile.

    Args:
        *children: The children to pass to the component.
        **props: The props to pass to the component.

    Returns:
        The component.
    """
    return Box.create(*children, **props, display=["block", "none", "none", "none"])


def tablet_only(*children, **props):
    """Create a component that is only visible on tablet.

    Args:
        *children: The children to pass to the component.
        **props: The props to pass to the component.

    Returns:
        The component.
    """
    return Box.create(*children, **props, display=["none", "block", "block", "none"])


def desktop_only(*children, **props):
    """Create a component that is only visible on desktop.

    Args:
        *children: The children to pass to the component.
        **props: The props to pass to the component.

    Returns:
        The component.
    """
    return Box.create(*children, **props, display=["none", "none", "none", "block"])


def tablet_and_desktop(*children, **props):
    """Create a component that is only visible on tablet and desktop.

    Args:
        *children: The children to pass to the component.
        **props: The props to pass to the component.

    Returns:
        The component.
    """
    return Box.create(*children, **props, display=["none", "block", "block", "block"])


def mobile_and_tablet(*children, **props):
    """Create a component that is only visible on mobile and tablet.

    Args:
        *children: The children to pass to the component.
        **props: The props to pass to the component.

    Returns:
        The component.
    """
    return Box.create(*children, **props, display=["block", "block", "block", "none"])



"""Statistics components."""

from pynecone.components.component import Component
from pynecone.components.libs.chakra import ChakraComponent
from pynecone.var import Var


class Stat(ChakraComponent):
    """The Stat component is used to display some statistics. It can take in a label, a number and a help text."""

    tag = "Stat"

    @classmethod
    def create(
        cls, *children, label=None, number=0, help_text=None, arrow_type=None, **props
    ) -> Component:
        """Create a stat component.

        Args:
            children: The children of the component.
            label: A label for the stat component.
            number: The value of the stat component.
            help_text: A text added to the stat component.
            arrow_type: The type of the arrow ("increase", "decrease", None)
            props: The properties of the component.

        Returns:
            The stat component.
        """
        if len(children) == 0:
            children = []
            if label:
                children.append(StatLabel.create(label))

            children.append(StatNumber.create(number))

            if help_text:
                if arrow_type:
                    children.append(
                        StatHelpText.create(
                            help_text, StatArrow.create(type_=arrow_type)
                        )
                    )
                else:
                    children.append(StatHelpText.create(help_text))

        return super().create(*children, **props)


class StatLabel(ChakraComponent):
    """A stat label component."""

    tag = "StatLabel"


class StatNumber(ChakraComponent):
    """The stat to display."""

    tag = "StatNumber"


class StatHelpText(ChakraComponent):
    """A helper text to display under the stat."""

    tag = "StatHelpText"


class StatArrow(ChakraComponent):
    """A stat arrow component indicating the direction of change."""

    tag = "StatArrow"

    # The type of arrow, either increase or decrease.
    type_: Var[str]


class StatGroup(ChakraComponent):
    """A stat group component to evenly space out the stats."""

    tag = "StatGroup"



"""Table components."""

from pynecone.components.component import Component
from pynecone.components.layout.foreach import Foreach
from pynecone.components.libs.chakra import ChakraComponent
from pynecone.var import Var


class Table(ChakraComponent):
    """A table component."""

    tag = "Table"

    # The color scheme of the table
    color_scheme: Var[str]

    # The variant of the table style to use
    variant: Var[str]

    # The size of the table
    size: Var[str]

    # The placement of the table caption.
    placement: Var[str]

    @classmethod
    def create(
        cls, *children, caption=None, headers=None, rows=None, footers=None, **props
    ) -> Component:
        """Create a table component.

        Args:
            children: The children of the component.
            caption: The caption of the table component.
            headers: The headers of the table component.
            rows: The rows of the table component.
            footers: The footers of the table component.
            props: The properties of the component.

        Returns:
            The table component.
        """
        if len(children) == 0:
            children = []

            if caption:
                children.append(TableCaption.create(caption))

            if headers:
                children.append(Thead.create(headers=headers))

            if rows:
                children.append(Tbody.create(rows=rows))

            if footers:
                children.append(Tfoot.create(footers=footers))
        return super().create(*children, **props)


class Thead(ChakraComponent):
    """A table header component."""

    tag = "Thead"

    @classmethod
    def create(cls, *children, headers=None, **props) -> Component:
        """Create a table header component.

        Args:
            children: The children of the component.
            props: The properties of the component.
            headers (list, optional): List of headers. Defaults to None.

        Returns:
            The table header component.
        """
        if len(children) == 0:
            children = [Tr.create(cell_type="header", cells=headers)]
        return super().create(*children, **props)


class Tbody(ChakraComponent):
    """A table body component."""

    tag = "Tbody"

    @classmethod
    def create(cls, *children, rows=None, **props) -> Component:
        """Create a table body component.

        Args:
            children: The children of the component.
            props: The properties of the component.
            rows (list[list], optional): The rows of the table body. Defaults to None.

        Returns:
            Component: _description_
        """
        if len(children) == 0:
            children = [Tr.create(cell_type="data", cells=row) for row in rows or []]
        return super().create(*children, **props)


class Tfoot(ChakraComponent):
    """A table footer component."""

    tag = "Tfoot"

    @classmethod
    def create(cls, *children, footers=None, **props) -> Component:
        """Create a table footer component.

        Args:
            children: The children of the component.
            props: The properties of the component.
            footers (list, optional): List of footers. Defaults to None.

        Returns:
            The table footer component.
        """
        if len(children) == 0:
            children = [Tr.create(cell_type="header", cells=footers)]
        return super().create(*children, **props)


class Tr(ChakraComponent):
    """A table row component."""

    tag = "Tr"

    @classmethod
    def create(cls, *children, cell_type: str = "", cells=None, **props) -> Component:
        """Create a table row component.

        Args:
            children: The children of the component.
            props: The properties of the component.
            cell_type (str): the type of cells in this table row. "header" or "data". Defaults to None.
            cells (list, optional): The cells value to add in the table row. Defaults to None.

        Returns:
            The table row component
        """
        types = {"header": Th, "data": Td}
        cell_cls = types.get(cell_type)
        if len(children) == 0 and cell_cls:
            if isinstance(cells, Var):
                children = [Foreach.create(cells, cell_cls.create)]
            else:
                children = [cell_cls.create(cell) for cell in cells or []]
        return super().create(*children, **props)


class Th(ChakraComponent):
    """A table header cell component."""

    tag = "Th"

    # Aligns the cell content to the right.
    is_numeric: Var[bool]


class Td(ChakraComponent):
    """A table data cell component."""

    tag = "Td"

    # Aligns the cell content to the right.
    is_numeric: Var[bool]


class TableCaption(ChakraComponent):
    """A table caption component."""

    tag = "TableCaption"

    # The placement of the table caption. This sets the `caption-side` CSS attribute.
    placement: Var[str]


class TableContainer(ChakraComponent):
    """The table container component renders a div that wraps the table component."""

    tag = "TableContainer"



"""Display the title of the current page."""


from pynecone.components.component import Component
from pynecone.var import Var


class Link(Component):
    """A component that displays the title of the current page."""

    tag = "link"

    # The href.
    href: Var[str]

    # The type of link.
    rel: Var[str]


class ScriptTag(Component):
    """A script tag with the specified type and source."""

    tag = "script"

    # The type of script represented.
    type_: Var[str]

    # The URI of an external script.
    source: Var[str]

    # Metadata to verify the content of the script.
    integrity: Var[str]

    # Whether to allow cross-origin requests.
    crossorigin: Var[str]

    # Indicates which referrer to send when fetching the script.
    referrer_policy: Var[str]

    # Whether to asynchronously load the script.
    is_async: Var[bool]

    # Whether to defer loading the script.
    defer: Var[bool]



"""Dynamically compile classes for all HTML elements and output them to the
elements directory.

This script generates the element classes in the pynecone.el.elements module.
Run as follows:

    python -m pynecone.el.precompile

Make sure to delete the __init__.py file in the elements directory before
running this script.
"""

import os

from pynecone.utils import path_ops

from .constants import ELEMENT_TO_PROPS, ELEMENTS

FILE_DIR = os.path.dirname(os.path.realpath(__file__))
ELEMENTS_DIR = os.path.join(FILE_DIR, "elements")
INIT_PY_PATH = os.path.join(ELEMENTS_DIR, "__init__.py")


def element_path(element: str) -> str:
    """Get the name of the Python file for the given element.

    Args:
        element: The name of the element. For example, `a` or `div`.

    Returns:
        The name of the Python file for the given element.
    """
    return os.path.join(ELEMENTS_DIR, f"{element}.py")


PROP = "    {prop}: PCVar[Union[str, int, bool]]".format


def compile_pyclass_props(element: str) -> str:
    """Compile props for an element.

    Args:
        element: The name of the element. For example, `a` or `div`.

    Returns:
        A string containing compiled props for the element.
    """
    return path_ops.join(PROP(prop=prop) for prop in ELEMENT_TO_PROPS[element])


PYCLASS = path_ops.join(
    [
        "",
        "class {name}(Element):  # noqa: E742",
        '    """Display the {element} element."""',
        "",
        '    tag = "{element}"',
        "",
        "{props}",
        "",
        "",
        "{call_name} = {name}.create",
        "",
    ]
).format


def compile_pyclass(element: str) -> str:
    """Compile a Python class for an element.

    Args:
        element: The name of the element. For example, `a` or `div`.

    Returns:
        A string containing a Python class for the element.
    """
    name = element.capitalize()
    props = compile_pyclass_props(element)

    # Handle the `del` element, which is a Python keyword. Note that the class
    # name is still `Del`.
    call_name = "del_" if element == "del" else element

    return PYCLASS(
        name=name,
        element=element,
        props=props,
        call_name=call_name,
    )


INIT_PY = [
    '"""Element classes. This is an auto-generated file. Do not edit. See ../generate.py."""',
    "from typing import Union",
    "",
    "from pynecone.el.element import Element",
    "from pynecone.var import Var as PCVar",
    "",
]


for element in sorted(ELEMENTS):
    INIT_PY.append(compile_pyclass(element))


os.makedirs(ELEMENTS_DIR, exist_ok=True)
with open(INIT_PY_PATH, "w+") as f:
    f.write(path_ops.join(INIT_PY))



"""Middleware to hydrate the state."""
from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, Union

from pynecone import constants
from pynecone.event import Event, EventHandler, get_hydrate_event
from pynecone.middleware.middleware import Middleware
from pynecone.state import State, StateUpdate
from pynecone.utils import format

if TYPE_CHECKING:
    from pynecone.app import App


IS_HYDRATED = "is_hydrated"


State.add_var(IS_HYDRATED, type_=bool, default_value=False)


class HydrateMiddleware(Middleware):
    """Middleware to handle initial app hydration."""

    async def preprocess(
        self, app: App, state: State, event: Event
    ) -> Optional[Union[StateUpdate, List[StateUpdate]]]:
        """Preprocess the event.

        Args:
            app: The app to apply the middleware to.
            state: The client state.
            event: The event to preprocess.

        Returns:
            An optional delta or list of state updates to return.
        """
        if event.name == get_hydrate_event(state):
            route = event.router_data.get(constants.RouteVar.PATH, "")
            if route == "/":
                load_event = app.load_events.get(constants.INDEX_ROUTE)
            elif route:
                load_event = app.load_events.get(route.lstrip("/"))
            else:
                load_event = None

            updates = []

            # first get the initial state
            delta = format.format_state({state.get_name(): state.dict()})
            if delta:
                updates.append(StateUpdate(delta=delta))

            # then apply changes from on_load event handlers on top of that
            if load_event:
                if not isinstance(load_event, List):
                    load_event = [load_event]
                for single_event in load_event:
                    updates.append(
                        await self.execute_load_event(
                            state, single_event, event.token, event.payload
                        )
                    )
            # extra message telling the client state that hydration is complete
            updates.append(
                StateUpdate(
                    delta=format.format_state({state.get_name(): {IS_HYDRATED: True}})
                )
            )

            return updates

    async def execute_load_event(
        self, state: State, load_event: EventHandler, token: str, payload: Dict
    ) -> StateUpdate:
        """Execute single load event.

        Args:
            state: The client state.
            load_event: A single load event to execute.
            token: Client token
            payload: The event payload

        Returns:
            A state Update.

        Raises:
            ValueError: If the state value is None.
        """
        substate_path = format.format_event_handler(load_event).split(".")
        ex_state = state.get_substate(substate_path[:-1])
        if not ex_state:
            raise ValueError(
                "The value of state cannot be None when processing an on-load event."
            )

        return await state._process_event(
            handler=load_event, state=ex_state, payload=payload, token=token
        )
